unit Database; {Fractal Landscapes 3.0 - Copyright  1987..1997, Jon Shemitz}

{Database and landscape generation}

interface

uses SysUtils, Global;

{ Misc math fns and such }

function IDIV(Numerator: LongInt; Denominator: Int16): Int16;
{$ifdef Ver80} {Delphi 1.0 still supported InLine fns}
InLine(
         $5B /                  { POP   BX ; Denominator    }
         $58 /                  { POP   AX ; Lo word of Num }
         $5A /                  { POP   DX ; Hi word of Num }
         $F7 / $FB              { IDIV  BX ; Mixed DIV      }
      );
{$endif}

function IMUL(A, B: Int16): LongInt;
{$ifdef Ver80} {Delphi 1.0 still supported InLine fns}
InLine(
         $5B /                  { POP   BX }
         $58 /                  { POP   AX }
         $F7 / $EB              { IMUL  BX }
      );
{$endif}

function Rand(Envelope: integer): integer;

{database}

procedure ResetDB;

function GetTriple(const V: TVertex): TTriple; { DB[V] }

procedure SwapTriples(var A, B: TTriple);

function Midpoint(A, B: TVertex): TVertex;

function LoadLandscape(const FileName: TFileName): boolean;

function SaveLandscape(const FileName: TFileName): boolean;

{calculation}

procedure FractureTriangle(const A, B, C: TVertex; Plys: word);

function Unscale(ScaledCoordinate: LongInt): TCoordinate;
{$ifdef Ver80} {Delphi 1.0 still supported InLine fns}
InLine(
         $58 /                    { POP  AX              ; Lo word of SC    }
         $5A /                    { POP  DX              ; Hi word of SC    }
         $8B / $1E / UnitLength / { MOV  BX,[UnitLength] ; Lo word of scale }
         $F7 / $FB                { IDIV BX              ; Unscale          }
      );
{$endif}

implementation

{ Misc math fns and such }

{$ifNdef Ver80} {32-bit Delphi doesn't support InLine fns}
function IDIV(Numerator: LongInt; Denominator: Int16): Int16;
begin
  Result := Numerator div Denominator;
end;
{$endif}

{$ifNdef Ver80} {32-bit Delphi doesn't support InLine fns}
function IMUL(A, B: Int16): LongInt;
begin
  Result := Longint(A) * B;
end;
{$endif}

function Rand(Envelope: integer): integer;
{ Pseudonormal distribution, in range Envelope }
begin
  Rand := integer(Random(Envelope)) + integer(Random(Envelope)) - Envelope;
end;

{$ifNdef Ver80} {32-bit Delphi doesn't support InLine fns}
function Unscale(ScaledCoordinate: LongInt): TCoordinate;
begin
  Result := ScaledCoordinate div UnitLength;
end;
{$endif}

{ Database }

var
  DB: array[0..8384] of TTriple; { Triangular array: Vertices(MEL+1) entries }

  NumberOfVertices, TopRow: word;

  Envelopes:  array[1..MaxPlys] of word;

function Vertices(N: word): word;
{ Vertices in an equilateral triangle with edgelength = N-1 }
begin
  Vertices := (Sqr(N) + N) shr 1;
end;

function Midpoint(A, B: TVertex): TVertex;
begin
  Result := Vertex( (A.AB + B.AB) shr 1, { Average }
                    (A.BC + B.BC) shr 1,
                    (A.CA + B.CA) shr 1 );
end;

function Loc(const V: TVertex): word;
begin
  Loc := NumberOfVertices - Vertices(TopRow - V.AB) + V.BC;
  {      ^^^^^^^^^^^^^^^^^^ This is actually NOT necessary and just
                            wastes cycles, but I have retained it
                            for compatability with FL2 .FL files. }

end;

procedure SetTriple(var V: TVertex; var T: TTriple);     { DB[V] := T }
begin
  DB[Loc(V)] := T;
end;

function GetTriple(const V: TVertex): TTriple; { DB[V] }
begin
  Result := DB[Loc(V)];
end;

procedure SwapTriples(var A, B: TTriple);
var
  Tmp: TTriple;
begin
  Tmp := A; A := B; B := Tmp;
end;

procedure SwapZ(var A, B: TTriple);
var
  C: TCoordinate;
begin
  C := A.Z; A.Z := B.Z; B.Z := C;
end;

const
  Uninitialized = -30000;

procedure ResetDB;
var
  T:          TTriple;
  R, Theta:   double;
  I, Offset:  integer;
  tA, tB, tC: TTriple;
const
  Base_Rotation = - Pi / 2.1; {Rotate point counterclockwise a bit}
  RotateBy      = Pi * 2 / 3; {120}
begin
  { Set Plys dependent stuff }
  EdgeLength := 1 shl (Plys - 1);
  TopRow := EdgeLength + 1; { A "fencepost" situation }
  NumberOfVertices := Vertices(TopRow);
  for I := Plys downto 1 do
    Envelopes[I] := Envelope shr Succ(Plys - I);
  { Then reset NumberOfVertices vertices in DB }
  T.X := Uninitialized;
  T.Y := Uninitialized;
  T.Z := Uninitialized;
  for I := Low(DB) to High(DB) do DB[I] := T;
  { Now, set "defining" (outside) points A, B, and C }
  A.AB := 0;          A.BC := EdgeLength; A.CA := 0;
  B.AB := 0;          B.BC := 0;          B.CA := EdgeLength;
  C.AB := EdgeLength; C.BC := 0;          C.CA := 0;
  { Then, assign them triples }
  Offset := UnitLength div 2;
  R      := UnitLength / 2;

  Theta := Base_Rotation;
  tA := Triple( Round(R * Cos(Theta)) + Offset,
                Round(R * Sin(Theta)) + Offset,
                SeaLevel + Rand(Envelope) );

  Theta := Theta + RotateBy;
  tB := Triple( Round(R * Cos(Theta)) + Offset,
                Round(R * Sin(Theta)) + Offset,
                SeaLevel + Rand(Envelope) );

  Theta := Theta + RotateBy;
  tC := Triple( Round(R * Cos(Theta)) + Offset,
                Round(R * Sin(Theta)) + Offset,
                SeaLevel + Rand(Envelope) );

  { At least one point above sealevel }
  if (tA.Z < SeaLevel) AND (tB.Z < SeaLevel) AND (tC.Z < SeaLevel) then
    repeat
      tB.Z := SeaLevel + Rand(Envelope);
    until tB.Z > SeaLevel;
  { Force A the lowest ... }
  if tA.Z > tB.Z then SwapZ(tA, tB);
  if tA.Z > tC.Z then SwapZ(tA, tC);

  SetTriple(A, tA);
  SetTriple(B, tB);
  SetTriple(C, tC);
end;

function SaveLandscape(const FileName: TFileName): boolean;
var
  Handle: integer;
begin
  try
    Handle := FileCreate(FileName);
    try
      Result := (FileWrite(Handle, Plys, SizeOf(Plys)) = SizeOf(Plys)) and
                (FileWrite(Handle, DB, NumberOfVertices * SizeOf(TTriple))
                 = NumberOfVertices * SizeOf(TTriple));
    finally
      FileClose(Handle);
    end;
  except
    on {any} Exception do Result := False;
  end;
end;

function LoadLandscape(const FileName: TFileName): boolean;
var
  Handle:  integer;
begin
  Result := False;
  try
    Handle := SysUtils.FileOpen(FileName, fmOpenRead);
    try
      if FileRead(Handle, Plys, SizeOf(Plys)) = SizeOf(Plys) then
        begin
        ResetDB;
        LoadLandscape := FileRead( Handle, DB,
                                   NumberOfVertices * SizeOf(TTriple))
                         = NumberOfVertices * SizeOf(TTriple);
        end;
    finally
      FileClose(Handle);
    end;
  except
    on {any} Exception do Result := False;
  end;
end;

{ Action }

procedure FractureLine( var vM: TVertex;
                        const vA, vB: TVertex;
                        Envelope: integer );
var
  A, B, M: TTriple;
begin
  vM := Midpoint(vA, vB);
  M  := GetTriple(vM);
  if M.X = Uninitialized then { Not set yet }
    begin
    A := GetTriple(vA); B := GetTriple(vB);
    M := Triple( A.X + (B.X - A.X) div 2,
                 A.Y + (B.Y - A.Y) div 2,
                 A.Z + (B.Z - A.Z) div 2 + Rand(Envelope) );
    { Mean height  Random(Envelope) }
    SetTriple(vM, M);
    end;
end;

procedure FractureTriangle(const A, B, C: TVertex; Plys: word);
var
  Envelope: word;
  AB, BC, CA: TVertex;
begin
  if Plys > 1 then
    begin
    Envelope := Envelopes[Plys];
    FractureLine(AB, A, B, Envelope);
    FractureLine(BC, B, C, Envelope);
    FractureLine(CA, C, A, Envelope);
    Dec(Plys);
    FractureTriangle(CA, BC, C, Plys);
    FractureTriangle(AB, B, BC, Plys);
    FractureTriangle(BC, CA, AB, Plys);
    FractureTriangle(A, AB, CA, Plys);
    end;
end;

end.



